"""Library to integrate with the Porsche Connect API.

This library provides a Python interface to Porsche Connect API, with
abilities to read vechicle status data and access to remote services
to control certain vechicle functions.

NOTE: This work is not officially supported by Porsche and functionality
can stop working at any time without warning.

"""
